;
;	Decode a track
;
;	1.	First find sync marks
;
;	Data	1 0 1 0 0 0 0 1|
;	Clock	 0 0 0 1 1 1 0 |
;	MFM		100010010x01001|
;
;	x missing clock for SYNC (0xA1) mark
;
;	To find a SYNC we need to find the sequence
;
;	LONG-NORM-LONG-NORM
;
;	2.	Find pattern
;
;	A normal diskette is formatted using certain patterns
;	to allow the controller to syncronise
;
;	Count	Pattern
;	12		0x00	SYNC
;					Data	0 0 0 0 0 0 0 0|
;					Clock	 1 1 1 1 1 1 1 |
;					MFM		010101010101010|
;	 3		0xA1	SYNC MARK
;
;					Data	1 0 1 0 0 0 0 1| 1 0 1 0 0 0 0 1| 1 0 1 0 0 0 0 1|
;					Clock	 0 0 0 1 1 1 0 |0 0 0 0 1 1 1 0 |0 0 0 0 1 1 1 0 |
;					MFM		100010010x01001|0100010010x01001|0100010010x01001|
;
;	 1		0xFE	Address Mark
;	or
;	 1		0xFB	Data Mark (also possible 0xF8 Deleted Data Mark)
;
;	 b		0xXX	Data
;
;	 2		CRC		16-bit cyclical redundancy check
;
;	So we should look for a certain amount of short gaps (up to 12*8)
;	followed by three SYNC MARKS, that is 6 groups of consecutive
;	long and medium gaps.
;	S	= short
;	M	= medium
;	L	= long
;	so it's landing on the following gap pattern
;	...SSSSSSSSSSSSSSSSSSSLMLMSLMLMSLMLM<start of data>
;
;	Timer constants are for a MCU clock of 18MHz and
;	HD 3 1/2 diskette transfer rate of 500kbps. Pulses
;	are separated by a 2 or more microseconds gap
;
;	short gap		2usec
;	medium gap		3usec
;	long gap		4usec
;
;	At 18.432MHz a short gap is equivalent to 37 cycles
;	a medium gap is 55 cycles and a long gap is 74
;	cycles. The mid point between the gap sizes are
;	(37+55)/2 = 46
;	(55+74)/2 = 64
;
;	The idea is to measure the time between the gaps, for this we
;	use the 8-bit counter 2 in free running mode without any outputs.
;	First the MCU waits for a falling edge of the RD signal and sets
;	the counter to 0. Then when it detects the next falling edge it
;	will retrieve the counter value which is a measure of the gap.
;
;
.equ	t_long	= 64
.equ	t_short	= 46

.equ	tracksize	= 12544			; Multiple of 256 and larger than 12500bytes net track size
.equ	trackbuffer	= 0x800			; where we place the buffer during tests
.equ	syncmarks	= 0x700			; where we place found sync marks

;
;	To simplify the code we use a macro that waits on a falling edge
;	of the RD signal, retrieves the counter and sets the counter to
;	0 for the next measurment

	.macro	btime
	sbis	PINC, RD
	rjmp	PC-1
	sbic	PINC, RD
	rjmp	PC-1
	lds		temp, TCNT2
	sts		TCNT2, zero
	.endmacro


;
;	Initialise
;
	ldi		xl, low(tracksize)		; Bytes to collect
	ldi		xh, high(tracksize)
	ldi		yl, low(trackbuffer)	; Buffer
	ldi		yh, high(trackbuffer)
	ldi		zl, low(syncmarks)		; List of SYNC Marks
	ldi		zh, high(syncmarks)

sstart:
	ldi		count, 80				; less than 12*8
sloop:
	btime
	cpi		temp, t_long
	brsh	sstart
	dec		count
	brne	sloop
;
;	Note typically there are 3 consecutive SYNC MARKS
;	LMLMSLMLMSLMLM
;
wlong1:
	btime
	cpi		temp, t_long
	brlo	wlong1
;
	btime
	cpi		temp, t_short
	brlo	wlong1
	cpi		temp, t_long
	brsh	wlong1
;
	btime
	cpi		temp, t_long
	brlo	wlong1
;
	btime
	cpi		temp, t_short
	brlo	wlong1
	cpi		temp, t_long
	brsh	wlong1
;
	btime
	cpi		temp, t_short
	brsh	wlong1
;
	btime
	cpi		temp, t_long
	brlo	wlong1
;
	btime
	cpi		temp, t_short
	brlo	wlong1
	cpi		temp, t_long
	brsh	wlong1
;
	btime
	cpi		temp, t_long
	brlo	wlong1
;
	btime
	cpi		temp, t_short
	brlo	wlong1
	cpi		temp, t_long
	brsh	wlong1
;
	btime
	cpi		temp, t_short
	brsh	wlong1
;	btime
	cpi		temp, t_long
	brlo	wlong1
;
	btime
	cpi		temp, t_short
	brlo	wlong1
	cpi		temp, t_long
	brsh	wlong1
;
	btime
	cpi		temp, t_long
	brlo	wlong1
;
	btime
	cpi		temp, t_short
	brlo	wlong1
	cpi		temp, t_long
	brsh	wlong1
;


;
;	Note typically there are 3 consecutive SYNC MARKS
;
wlong1:
	btime
	cpi		temp, t_long
	brlo	wlong1
;
;	We have seen a long gap between RD pulses
;
wnorm1:
	btime														;
	cpi		temp, t_short										;
	brlo	wlong1												;
	cpi		temp, t_long										;
	brsh	wnorm1												;
;
;	We have seen a medium gap after a long gap
;
	btime														;
	cpi		temp, t_long
	brlo	wlong1			; we need a long gap now
	btime
	cpi		temp, t_short
	brlo	wlong1
	cpi		temp, t_long
	brhs	wnorm1



;
;	we spotted  a sync mark now we need to analyze further gaps
;	there are several cases depending on the data following the
;	sync mark.
;
;	Next Bit is a 1
;	
;	Data	1 0 1 0 0 0 0 1| 1
;	Clock	 0 0 0 1 1 1 0 |0
;	MFM		100010010001001|01
;
	ldi		count, 8
l0000:															; Worst case must be less than 37 cycles (short gap)
	btime														;  8
	cpi		temp, t_short										;  1
	brsh	l0020												;  2	-> 11
	sec															;
	rol		char												;
	dec		count												;
	brne	l0000												;
	st		Y+, char											;
	ldi		count, 8											;
	sbiw	xh:xl, 1											;
	brne	l0000												;
	rjmp	readtrackdone										;

l0020:
	cp		temp, t_long										;  1
	brlo	l0040												;  2	-> 14
;---------------------------------------------------------
;
;	Data	1 0 1 0 0 0 0 1| 0 1
;	Clock	 0 0 0 1 1 1 0 |0 0
;	MFM		100010010001001|0001
;
	lsl		char												;
	dec		count												;
	brne	l0030												;
	st		Y+, char											;
	ldi		count, 8											;
	sbiw	xh:xl, 1											;
	brne	l0000												;
	rjmp	readtrackdone										;
l0030:
	sec															;
	rol		char												;
	dec		count												;
	brne	l0000												;
	st		Y+, char											;
	ldi		count, 8											;
	sbiw	xh:xl, 1											;
	brne	l0000												;
	rjmp	readtrackdone

l0040:
;---------------------------------------------------------
;
;			------------------->
;	Data	1 0 1 0 0 0 0 1| 0 0
;	Clock	 0 0 0 1 1 1 0 |0 1 
;	MFM		100010010001001|001
;
	lsl		char												;  1
	dec		count												;  1
	brne	l0050												;  2	-> 18
	st		Y+, char											;
	ldi		count, 8											;
	sbiw	xh:xl, 1											;
	brne	l0050												;
	rjmp	readtrackdone
;
l0050:
	lsl		char												;  1
	dec		count												;  1
	brne	l0060												;  1
	st		Y+, char											;  2
	ldi		count, 8											;  1
	sbiw	xh:xl, 1											;  2
	brne	l0060												;  2	-> 28
	rjmp	readtrackdone
;
;	pulse was on a clock and the next data bit is either
;	a zero or a one in which case the next gap is either
;	a short (zero) or a medium(one) a long gap at this
;	stage would be equivalent to a sync mark
;
l0060:
	bittime
	cpi		temp, t_short
	brlo	l0080
	cpi		temp, t_long
	brsh	l0090
;
;	we have a medium gap and therefore the data bit is one
;	and we are again aligned to the data bits
;
;			------------------->
;	Data	1 0 1 0 0 0 0 1| 0 0 1
;	Clock	 0 0 0 1 1 1 0 |0 1 0
;	MFM		100010010001001|001001
;
	sec
	rol		char
	dec		count
	brne	l0010
	st		Y+, char
	ldi		count, 8
	sbiw	xh:xl, 1
	brne	l0010
	rjmp	readtrackdone
;
l0080:
;
;	we have a short pulse, and therefore the data bit is
;	zero and we are back at the situation where the pulse
;	was on a clock
;
;			------------------->
;	Data	1 0 1 0 0 0 0 1| 0 0 0
;	Clock	 0 0 0 1 1 1 0 |0 1 1
;	MFM		100010010001001|00101
;
	lsl		char
	dec		count
	brne	l0060
	st		Y+, char
	ldi		count, 8
	sbiw	xh:xl, 1
	brne	l0060
;
l0090:
;
;	we have a long gap, at this stage this is equivalent
;	to a missing clock pulse, a long gap with missing
;	clock clock pulse means a medium gap and then again
,	data is expected
;
;			------------------->
;	Data	1 0 1 0 0 0 0 1| 0 0 0 0
;	Clock	 0 0 0 1 1 1 0 |0 1 0 1
;	MFM		100010010001001|0010001

	st		Z+, yl
	st		Z+, yh
	bittime
	rjmp	l0010
;
;
readtrackdone:


;
;
;
;
	analyze buffer
	
	
	ld		char, Y+
	clr		temp
	lsl		char
	rol		temp
	lsl		char
	rol		temp
	cpi		temp, c_short
	breq
	cpi		temp, c_medium
	breq
	cpi		temp, c_long
	
	;
	;	we spotted  a sync mark now we need to analyze further gaps
	;	there are several cases depending on the data following the
	;	sync mark.
	;
	b0000:
		ldi		count, 8
	b0010:															; Worst case must be less than 37 cycles (short gap)
		btime														;  8
		cpi		temp, t_short										;  1
		brsh	b0020												;  2	-> 11
	;---------------------------------------------------------
	;
	;	Next Bit is a '1'
	;			--------------->
	;	Data	1 0 1 0 0 0 0 1| 1
	;	Clock	 0 0 0 1 1 1 0 |0
	;	MFM		100010010001001|01
	;
		sec															;
		rol		char												;
		dec		count												;
		brne	b0010												;
		rjmp	gotbyteeven
	
	b0020:
		cpi		temp, t_long										;  1
		brlo	b0040												;  2	-> 14
	;---------------------------------------------------------
	;
	;	Next are two bits '01'
	;			--------------->
	;	Data	1 0 1 0 0 0 0 1| 0 1
	;	Clock	 0 0 0 1 1 1 0 |0 0
	;	MFM		100010010001001|0001
	;
		lsl		char												;
		dec		count												;
		brne	b0030												;
		rjmp	gotbyteeven1
	b0030:
		sec															;
		rol		char												;
		dec		count												;
		rjmp	gotbyteeven
	;---------------------------------------------------------
	;
	;	1 1/2 bit cells so we have a clock pulse and next two
	;	bits are '00'
	;			--------------->
	;	Data	1 0 1 0 0 0 0 1| 0 0
	;	Clock	 0 0 0 1 1 1 0 |0 1 
	;	MFM		100010010001001|001
	;
	b0040:
		lsl		char												;  1
		dec		count												;  1
		brne	b0050												;  2	-> 18
		rjmp	gotbyteodd0
	;
	b0050:
		lsl		char												;  1
		dec		count												;  1
		brne	b0060												;  1
		rjmp	gotbyteodd
	;---------------------------------------------------------
	;
	;	pulse was on a clock and the next data bit is either
	;	a zero or a one in which case the next gap is either
	;	a short (zero) or a medium(one) a long gap at this
	;	stage would be equivalent to a sync mark
	;
	b0060:
		btime
		cpi		temp, t_short
		brlo	b0080
		cpi		temp, t_long
		brsh	b0090
	;---------------------------------------------------------
	;
	;	we have a medium gap and therefore the data bit is one
	;	and we are again aligned to the data bits
	;
	;			------------------->
	;	Data	1 0 1 0 0 0 0 1| 0 0 1
	;	Clock	 0 0 0 1 1 1 0 |0 1 0
	;	MFM		100010010001001|001001
	;
		sec
		rol		char
		dec		count
		brne	b0010
		rjmp	gotbyteeven
	;---------------------------------------------------------
	;
	;	we have a short pulse, and therefore the data bit is
	;	zero and we are back at the situation where the pulse
	;	was on a clock
	;
	;			------------------->
	;	Data	1 0 1 0 0 0 0 1| 0 0 0
	;	Clock	 0 0 0 1 1 1 0 |0 1 1
	;	MFM		100010010001001|00101
	;
	b0080:
		lsl		char
		dec		count
		brne	b0010
		rjmp	gotbyteodd
	;---------------------------------------------------------
	;
	;	we have a long gap, at this stage this is equivalent
	;	to a missing clock pulse, a long gap with missing
	;	clock pulse means a medium gap and then again
	;	data is expected
	;
	;	Note: A long gap can normally not occur with the
	;	pulses synced to the clock position. Such a gap
	;	would require the the data bit after the first
	;	pulse is 0 and at the same time the next data bit
	;	which is right before the second pulse as well. 
	;	But two data bits produce a clock pulse in between
	;	which obviously is missing
	;
	;			------------------->
	;	Data	1 0 1 0 0 0 0 1| 0 0 0 0
	;	Clock	 0 0 0 1 1 1 0 |0 1 0 1
	;	MFM		100010010001001|0010001
	;
	b0090:
		st		Z+, yl
		st		Z+, yh
	;	rjmp	sstart
