;==========================================================================
;
;	Parsing a command line:
;
;	Command lines consist of keywords, strings, etc.. Keywords can be
;	abreviated and are case insensitive. Keywords may have a suffix
;	in the form of a number or a number followed by a colon. The parser
;	uses tables with 3 entries
;
;		-	pointer to desciption of syntax element, 0 means no further 
;			checks possible, i.e. the scan of the command failed or in
;			other words this is not a valid command
;		-	action routine to call in case of success, 0 no action routine
;			this is typically required to make note of the results produced
;			when scanning the syntax element, e.g. the suffix number, the 
;			number or the string/quoted string found
;		-	if the element matches then the third entry gives the address
;			of the next parser table or "cmd_done" in case the end of the
;			command has been reached, if no match the parser proceeds
;			to the following entry
;
;	A syntax element consists of a type followd by additional parameters. In
;	case of keywords the name of the keyword is given the mandatory part must
;	be given in capital and the optional part in lower case. In case of a sub-
;	command the parameter is a pointer to a command table (the parser calls 
;	itself using this pointer as the table pointer)
;
;	A Table must be terminated with a 0 to indicated the end. If you want
;	a command to be valid you must typically scan up to the terminating 
;	character, in our case <CR>, using the eol syntax element and a next
;	table pointer of cmd_done ( -1). You can use cmd_done with every syntax
; 	element type. 
;
;	
;
;	When calling the parser Z must point to the start table (caution it must
;	be a byte pointer for lpm, typically twice the address) and X must point
;	to the string to be parsed. If the sring could be parsed successfully
;	carry will be cleared on return. Else carry will be set. X will point
;	to the position up to which the parser has analyzed the string (not
;	necessarily the end of the string)
;
.equ	cmd_done =  -1


.macro	P_KEYWORD
.if (STRLEN(@0)&1)
.db	0x80, @0, 0x00, 0x00
.else
.db	0x80, @0, 0x00
.endif
.endmacro
;
.macro	P_KEYSUFF
.if (STRLEN(@0)&1)
.db	0x81, @0, 0x00, 0x00
.else
.db	0x81, @0, 0x00
.endif
.endmacro
;
.macro	P_STRING
.db		0x82, 0x00
.endmacro
;
.macro	P_KEYDEVICE
.if (STRLEN(@0)&1)
.db	0x83, @0, 0x00, 0x00
.else
.db	0x83, @0, 0x00
.endif
.endmacro
;
.macro	P_NAME
.db		0x84, 0x00
.endmacro
;
.macro	P_NUMBER
.db		0x85, 0x00
.endmacro
;
.macro	P_EOL
.db		0x86, 0x00
.endmacro
;
.macro	P_QSTRING
.db		0x87, 0x00
.endmacro
;
.macro	P_SUBCMD
.db		0x88, 0x00
.dw		@0
.endmacro
;
.macro	P_CHAR
.db		0x89, @0
.endmacro
;
.macro	P_ANY
.db		0x8a, 0x00
.endmacro
;
.macro	P_LAMBDA
.db		0x8b, 0x00
.endmacro
;--------------------------------------------------------------------------
;
;	Re-entering scancommand with subcommand
;
;	Input:
;		Z		Pointer to parser table in flash
;		X		Pointer to string
;
scansubcmd:
	call	print
	.db		0x0a, 0x0d, "Subcommand:", 0x0a, 0x0d, 0x00
	movw	zh:zl, r25:r24
	adiw	zh:zl, 1
	lpm		r24, Z+
	lpm		r25, Z+
	add		r24, r24
	adc		r25, r25
	movw	zh:zl, r25:r24
;--------------------------------------------------------------------------
;
;	Process Command Table Entry
;
;	Input:
;		Z		Pointer to parser table
;		X		Pointer to string
;
;	Output:
;		Z		
;		X		Points after the last byte successfully parsed
;
;	Conditions:
;		CS		syntax error
;		CC		parsing string sucessfully
;
scancommand:
;	add		zl, zl
;	adc		zh, zh
	lpm		r24, Z+
	lpm		r25, Z+					; Get address of syntax element
	cp		r24, zero
	cpc		r25, zero
	breq	scancommandfailed		; End of Table
	push	xl
	push	xh						; Remember pointer into buffer
	push	zl
	push	zh						; Save table pointer
	add		r24, r24
	adc		r25, r25				; Convert address to Z byte pointer
	movw	zh:zl, r25:r24			;
	rcall	scanskipblank
;	rcall	scanreport
	rcall	scanelement
	brcs	scannextentry
;	rcall	scanstatus
	pop		zh
	pop		zl						; Restore Table pointer
	lpm		r24, Z+
	lpm		r25, Z+
	cp		r24, zero
	cpc		r25, zero
	breq	scannoaction
	push	zl
	push	zh
	movw	zh:zl, r25:r24
	icall	
	pop		zh
	pop		zl
	brcs	scannextentry2
scannoaction:
	pop		r24
	pop		r24						; Discard saved buffer pointer
	lpm		r24, Z+
	lpm		r25, Z+					; Next Table
	cp		r24, zero
	cpc		r25, zero
	breq	scancommand				; Just next entry
	cp		r24, ff
	cpc		r25, ff
	breq	scancommanddone
	add		r24, r24
	adc		r25, r25
	movw	zh:zl, r25:r24
	rjmp	scancommand

scannextentry:
;	call	seroutcrlf
	pop		zh
	pop		zl						; Restore Table pointer
	pop		xh
	pop		xl						; Previous pointer
	adiw	zh:zl, 4				; Skip action and next table address
	rjmp	scancommand

scannextentry2:
;	call	seroutcrlf
	pop		xh
	pop		xl						; Restore buffer pointer
	adiw	zh:zl, 2				; Skip next table address
	rjmp	scancommand

;
;
;
scancommandfailed:
	call	print
	.db		0x0a, 0x0d, "Unidentified Command.", 0x0a, 0x0d, 0x00
	sec
	ret

scancommanddone:
	movw	zh:zl, r25:r24			
;	call	print
;	.db		"Valid Command. ", 0x0a, 0x0d, 0x00
	clc
	ret
;
;	Skip blanks
;
scanskipnext:
	adiw	X, 1
scanskipblank:
	ld		char, X
	cpi		char, ' '
	breq	scanskipnext
	ret
;
;	scanelement
;
;	X		pointer to inputbuffer
;	Z		pointer to element descriptione
;
scanelement:
	lpm		temp, Z+
	movw	r25:r24, zh:zl
	mov		zl, temp
	andi	zl, 0x7f					; Remove marker bit
	cpi		zl, scantablesize			; Check against table size
	brsh	scanillegal
	clr		zh							; Make pointer to jump table
	subi	zl, low(-scantable)
	sbci	zh, high(-scantable)
	ijmp								; Go

scanillegal:
	call	print
	.db		"Undefined Syntax Element!", 0x0a, 0x0d, 0x00
	sec
	ret

scantable:
	rjmp	scankeyword
	rjmp	scankeyword
	rjmp	scanstring
	rjmp	scankeyword
	rjmp	scanname
	rjmp	scannumber
	rjmp	scaneol
	rjmp	scanqstring
	rjmp	scansubcmd
	rjmp	scanchar
	rjmp	scanany
	rjmp	scanlambda
	
.equ scantablesize=PC-scantable

;
;
;
scankeyword:
	movw	zh:zl, r25:r24
	sts		scanresult, zero
	sts		scanresult+1, zero
	sts		scanresult+2, zero
	sts		scanresult+3, zero
	push	temp						; Remember options for suffix
scankeywordloo:
	ld		char, X+
	lpm		temp, Z+					; Get both characters
;
;	Case:	(Z) == 0x00
;
	tst		temp
	breq	scankeywordsuf				; Reach end of keyword
;
;	Case:	(Z) == Lowercase i.e. optional character in keyword
;
	cpi		temp, 0x60
	brlo	scankeywordupp
	andi	temp, 0x5f
	cpi		char, 0x60
	brlo	scankeyword010
	andi	char, 0x5f					; Translate input to upper case
scankeyword010:
	cp		char, temp
	breq	scankeywordloo				; Matching optional characters
	rjmp	scankeywordsuf				; Check for allowed suffix
;
;	Case:	(Z) == Uppercase i.e. required character in keyword
;
scankeywordupp:
	cpi		char, 0x60
	brlo	scankeyword020
	andi	char, 0x5f					; Translate input to upper case
scankeyword020:
	cp		temp, char
	breq	scankeywordloo				
	pop		temp
	sec									; No match for mandatory part
	ret
;
;	We have reached the end of the required keyword or we have reached
;	the optional part of the keyword and got a non-matching character.
;	Now we need to check for a suffix. If a suffix is allowed we either
;	have just a number (at the moment we accept only one digit) or an
;	empty string. 
;
scankeywordsuf:
	pop		temp						; Get options for suffix
	sbrs	temp, 0   					; 
	rjmp	scankeywordnos				; no suffix at all
	cpi		char, '0'
	brlo	scankeywordcol				; Check for a digit
	cpi		char, '9'+1
	brsh	scankeywordcol
	subi	char, '0'					; Convert digit to number
	sts		scanresult+2, char			; Save result
	ld		char, X+
scankeywordcol:							; Get next character
	sbrs	temp, 1   					; Check for colon
	rjmp	scankeywordnos				; No so we must have a whitespace
	cpi		char, ':'					; Must be colon
	brne	scankeyworderr
	ld		char, X+
scankeywordnos:							; no further suffix so the character
	cpi		char, 0x0d					; must be a white space
	breq	scankeywordsuc
	cpi		char, ' '
	breq	scankeywordsuc
scankeyworderr:
	sec									; word longer than keyword and no suffix allowed
	ret
scankeywordsuc:
	sbiw	X, 1
	clc
	ret
;
;	Scans an arbitrary string, i.e. any non-blank characters
;	the string must not be empty, returns the pointer and
;	the length in scanresult. 
;
scanstring:
	sts		scanresult, xl
	sts		scanresult+1, xh
	sts		scanresult+2, zero
	sts		scanresult+3, zero
	mov		count, ff				; Initialise -1
scanstringnext:
	inc		count					; 
	ld		char, X+
	cpi		char, 0x0d				; end-of-line?
	breq	scanstringeol			; finish
	cpi		char, ' '				; space
	brne	scanstringnext			; no get next
scanstringeol:
	sts		scanresult+2, count		; save length
	tst		count
	breq	scanstringempty
	sbiw	X, 1
;	sts		pprint, xl
;	sts		pprint+1, xh
;	call	print
;	.db		"ptr after string:0x", 0x81, 0x80, 0x0d, 0x0a, 0x00
	clc
	ret
scanstringempty:
	sec
	ret
;
;
;
scanqstring:
	ld		temp, X+
	cpi		temp, 0x0d
	breq	scanqstringerr
	cpi		temp, '"'
	breq	scanqstringq
	cpi		temp, '\''
	breq	scanqstringq
	sec
	ret
scanqstringq:
	sts		scanresult, xl
	sts		scanresult+1, xh
	sts		scanresult+2, zero
	sts		scanresult+3, zero
	mov		count, ff

scanqstringnext:
	inc		count
	ld		char, X+
	cpi		char, 0x0d
	breq	scanqstringerr
	cp		char, temp
	brne	scanqstringnext
	sts		scanresult+2, count
	clc
	ret
scanqstringerr:
	sec
	ret
;
;	Makes sure we are at end of line
;
scaneol:
	sts		scanresult, zero
	sts		scanresult+1, zero
	sts		scanresult+2, zero
	sts		scanresult+3, zero
	ld		char, X
	cpi		char, 0x0d			; Clears carry if equal
	breq	scaneolret
	sec
scaneolret:
	ret

;
;	Scans a valid name. Names are string that start with a A-Z
;	and contain A-Z, 0-9, _, $, @
;
scanname:
	sts		scanresult, xl
	sts		scanresult+1, xh
	sts		scanresult+2, zero
	sts		scanresult+3, zero
	mov		count, one				; Assume we found at least one
	ld		char, X+				; character
	cpi		char, 0x60				; Conver to upper case
	brlo	scanname010
	andi	char, 0x5f
scanname010:
	cpi		char, 'A'				; Initial character A-Z?
	brlo	scannameinv
	cpi		char, 'Z'+1
	brsh	scannameinv		
scannamenext:					
	ld		char, X+				; Convert to upper case
	cpi		char, 0x60
	brlo	scanname020
	andi	char, 0x5f
scanname020:
	cpi		char, '$'				; Special Characters
	breq	scannameok
	cpi		char, '_'
	breq	scannameok
	cpi		char, '@'
	breq	scannameok
	cpi		char, '0'				; 
	brlo	scannamechk
	cpi		char, '9'+1
	brlo	scannameok				; Is between 0-9
scanname030:
	cpi		char, 'A'
	brlo	scannamechk
	cpi		char, 'Z'+1
	brsh	scannamechk
scannameok:							; Is between A-Z
	inc		count					; One more found
	rjmp	scannamenext			; and check next
scannamechk:
	cpi		char, 0x0d				; When no valid character
	breq	scannamesuc				; was found then we either
	cpi		char, ' '				; have to be at end of line
	breq	scannamesuc				; or a space terminates the name
scannameinv:
	sec								; Not the case -> error
	ret
scannamesuc:
	sbiw	X, 1
	sts		scanresult+2, count		; Save length of name
	clc								; Return -> sucess
	ret
;
;	A valid number
;
;	0nnnnn		Octal
;	0xnnnn		Hexadecimal
;	0bnnnn		Binary
;	nnnnnn		Decimal
;	nnnnnn.		Decimal, Overwrites Octal
;
;
scannumber:
	sts		scanresult, zero
	sts		scanresult+1, zero
	sts		scanresult+2, zero
	sts		scanresult+3, zero

	ld		char, X+
	cpi		char, 0x0d
	breq	scannumbererr
	
	ldi		temp, 10			; Assume Decimal
	cpi		char, '0'
	brne	scannumberdigit
	ldi		temp, 8				; Assume Octal
	ld		char, X+
	cpi		char, 0x0d
	breq	scannumbererr
	cpi		char, 0x60
	brlo	scannumber010
	andi	char, 0x5f
scannumber010:
	cpi		char, 'X'
	breq	scannumberhex
	cpi		char, 'B'
	breq	scannumberbinary
scannumberdigit:
	cpi		char, '9'+1
	brsh	scannumbererr
	subi	char, '0'
	cp		char, temp
	brsh	scannumbererr
	ret

scannumberhex:
scannumberbinary:
	ret
scannumbererr:
	sec
	ret
;
;	Match a specific character
;
scanchar:
	ld		char, X+
	cpi		char, 0x0d
	breq	scancharerr
	tst		char
	breq	scancharerr
	movw	zh:zl, r25:r24
	lpm		temp, Z
	cp		char, temp
	brne	scancharerr
	sts		scanresult+0, xl
	sts		scanresult+1, xh
	sts		scanresult+2, char
	clc
	ret
scancharerr:
	sec
	ret
;
;	Match any non-terminating character
;
scanany:
	ld		char, X+
	cpi		char, 0x0d
	breq	scananyerr
	tst		char
	breq	scananyerr
	sts		scanresult+2, char
	clc
	ret
scananyerr:
	sec
	ret
;
;	Match always
;
scanlambda:
	clc
	ret
